#ifndef ANALYSISSERVICE_H
#define ANALYSISSERVICE_H
#include <is/isinfo.h>
#include <ipc/object.h>
#include <ipc/server.h>

#include <string>
#include <boost/utility.hpp>
#include <boost/thread.hpp>

#include "SctFitter/FitStrategy.h"
#include "AnalysisServiceInterface.hh"
#include "AnalysisArguments.h"

namespace SctAnalysis {
class DcsInterface;
class AnalysisWorkerGroup;

class AnalysisService : 
  public IPCNamedObject<POA_AnalysisServiceI::AnalysisServiceInterface>,
  private boost::noncopyable 
  {
private:
    void shutdown();
 public:
    virtual ~AnalysisService();
    /** 
	RUNS the IPC server and IS callback. Does not return. 
    */
    void run();
    /**
       Callback for any data type
     */
    static void generalCallback(ISCallbackInfo* isc);
    /** 
	what to do when a callback is requestedfound, or removed 
    */
    static void testDataCallback(ISCallbackInfo * isc);
    /** 
	what to do when a new FitScanResult or RawScanResult is found 
    */
    static void scanResultCallback(ISCallbackInfo * isc);
    /** 
	set the FitStrategy by name 
    */
    void setFitStrategy(const string& name) throw(LogicError);
    /** 
	return a FitStrategy which tells how to do the fitting. 
    */ 
    SctFitter::FitStrategy& getFitStrategy() const throw(LogicError);
    /** 
	singleton access method 
    */
    static AnalysisService& instance();
    
    /**
      Initializes the static instance with a FitStrategy.  The default
      FitStrategy is RootFitStrategy.  Will throw ConfigurationException if something goes wrong.
      */
    static AnalysisService& initialize(AnalysisArguments args);
    /**
       Calls AnalysisWorkerGroup::purge() method to free stuck memory.
       @note Don't even think of doing this while analysis is in progress!
    */
    virtual void purge() throw();
    /**  
	 Analyse all data relevant to a test.
	 @param testname the TestData object IO name
    */
    virtual void analyze(const char* testname) throw();
    /**  
	 Analyse all data relevant to a test for a particular module.
	 @param testname the TestData object IO name
	 @param modulename the name of the module to be analyzed
    */
    virtual void analyzeModule(const char* testname, 
			       const char* modulename) throw();
    /**
       Output the status of the AnalysisService over IPC.
    */
    virtual char* status() throw();
    /** 
	ipc busy method 
     */
    virtual CORBA::Short busy ();
    /** 
	ipc queueLength method 
     */
    virtual CORBA::Short queueLength ();
    /**
	return the DcsInterface, which retrieves the dcs information 
    */
    DcsInterface& getDcsInterface();
    /**
       server to publish to. set by command line arguments
     */
    const std::string getOutputISServerName();
    /**
       recover from failure
    */
    void recover();

    static IPCServer& getServer() throw();
    
protected:
    /** 
	AnalysisWorkerGroup does the internal book-keeping, and calls the appropriate AnalysisAlgorithm 
    */
    AnalysisWorkerGroup* workergroup;
    DcsInterface* dcsinterface;
private:
    boost::recursive_mutex m_status_access ; ///< protects status from different threads.
    SctFitter::FitStrategy* fitStrategy; ///< cache telling how to do fitting.
    std::auto_ptr<ISInfoReceiver> infoReceiver;     ///<IS InfoReceiver for IS events
    static AnalysisService* service;
    AnalysisService(AnalysisArguments args); ///< singleton.  Will throw ConfigurationException if something goes wrong.
    AnalysisArguments m_args;
};
}

#endif //#ifndef ANALYSISSERVICE_H
