////////////////////////////////////////////////////////////////////////
//	    sct_is_server.cc
//
//	    IS binary
//
//	    description:
//		     Implements server functionality for the Information Service
//		
////////////////////////////////////////////////////////////////////////

#include <signal.h>
#include <fstream>

#include <cmdl/cmdargs.h>
#include <ipc/core.h>
#include <ipc/alarm.h>

#include "is/isrepository.h"
#include "is/exception.h"

#include "pmg/pmg_initSync.h"

#include "Sct/ApplicationStartupDebugTools.h"
#include "Sct/MultiMessageDebugStream.h"
#include "SubscriptionManager.hh"

static std::string	backupTo;
static ISRepository *	Repository = 0;
static int		signalReceived;

static bool backup( void * rock )
{
    ISRepository * f = static_cast<ISRepository*>( rock );
    
    std::cout << OWLTime() << " :: IS Server: Saving information ... " ;
    std::ofstream fout( backupTo.c_str(), std::ios::out | std::ios::trunc );
    if ( fout.good() )
    {
	fout << *f;
    }
    else
    {
	std::cerr << OWLTime()  << " :: WARNING [IS Server::backup()] Bad backup file - " 
				<< backupTo << std::endl;
    }
    std::cout << "done." << std::endl;    
    return true;
}

void signal_handler( int sig )
{
    if ( Repository != 0 )
    {
    	signalReceived = sig;
	Repository -> stop();
    }
}

int main(int argc, char ** argv)
{
    std::list< std::pair< std::string, std::string > > opt = IPCCore::extractOptions(argc,argv);
    
    opt.push_front( std::make_pair( std::string("threadPerConnectionPolicy"), std::string("0") ) );
    opt.push_front( std::make_pair( std::string("maxServerThreadPoolSize"),   std::string("3") ) );
    opt.push_front( std::make_pair( std::string("threadPoolWatchConnection"), std::string("0") ) );
    
    IPCCore::init( opt );
    Sct::ApplicationStartupDebugTools::announceStartOfMain(argc, argv, __FILE__, __LINE__);
    	
    CmdArgInt	period ('t', "time", "seconds","delay between backup operations");
    CmdArgStr	partition_name('p', "partition", "partition-name", "partition to work in.");
    CmdArgStr	server_name('n', "server", "server-name", "server to start.",CmdArg::isREQ);
    CmdArgStr	restore_from('r', "restore", "restore-from-file", "backup file to restore from.");
    CmdArgStr	backup_to('b', "backup", "backup-to-file", "file to backup to.");
    CmdArgInt	verbosity('v', "verbose", "verbosity-level", "possible levels are 0 (default), 1 and 2.");
    CmdArgInt	history_size('H', "history", "history-size", "set maximum number of values which will "
    							     "be kept for an information object.");
    CmdArgBool synchRequested('S',"synch", "Enable synchronisation using pmg_initSync()");
    CmdArgBool lookForSM('L',"look", "Look for SubscriptionManager.");
    CmdArgBool contIfSMNotFound('C',"cont", "Continue to run even in SubscriptionManager not foound (default=die).  Ignored if -L not specified.");

    CmdLine  cmd(*argv, &period, &partition_name, &server_name, &restore_from,
                 &backup_to, &verbosity, &history_size, &synchRequested,
                 &lookForSM, &contIfSMNotFound, NULL);
    CmdArgvIter	arg_iter(--argc, ++argv);

    cmd.description("This program implements server functionality for the Information Service.\n");
 
    period = -1;
    verbosity = 0;
    history_size = 100;
 
    // Parse arguments
    
    cmd.parse(arg_iter);
		
    if ( history_size < 1 )
    {
	std::cerr << OWLTime() << " :: ERROR [IS Server::main()] history size value is too small: " 
				<< history_size << std::endl;
	return 1;
    }
    
    IPCPartition     p(partition_name);

    if ( ISRepository::isExist( p, server_name ) )
    {
	std::cerr << OWLTime() << " :: ERROR [IS Server::main()] IS server " 
				<< server_name << " already running in the partition " 
				<< p.name() << std::endl;
	return 1;
    }

    signal(SIGINT,signal_handler);
    signal(SIGTERM,signal_handler);

    if ( backup_to.isNULL() )
    {
	backupTo = p.name();
	backupTo += ".";
	backupTo += server_name;
	backupTo += ".infoserver.backup";
    }
    else
    {
	backupTo = backup_to;
    }

    Repository = new ISRepository( p, server_name, history_size, verbosity ); 
    
    IPCAlarm * alarm = 0;
    if ( period > 0 )
    {
	alarm = new IPCAlarm( period, backup, Repository );
    }
	    
    if ( restore_from.isNULL() != true )
    {
	std::cout << OWLTime() << " :: IS Server: Restoring information from the " 
				<< restore_from << " file ... " << std::endl;
	std::ifstream fin(restore_from);
	if ( !fin.good() )
	{
	    std::cerr << OWLTime() << " :: ERROR [IS Server::main()] Can not read backup file " 
	    	      << restore_from << "." << std::endl;
	}
	else
	{
	    try 
	    {
		fin >> *Repository;
	    }
	    catch( is::BadBackupFile & ex )
	    {
		std::cerr << OWLTime() << " :: ERROR [IS Server::main()] Bad backup file " 
			  << restore_from << "." << std::endl;
	    }
	}
	std::cout << "done." << std::endl;
    }

    if (lookForSM) {
      const std::string helpMessage = "If you never intended to run a SubscriptionManager, then run without -L, though of course you will lose some functionality. If it is not important whether connection succeeds, add -C.  See 'sct_is_server --help' for more info.";
      try {
        Sct::MultiMessageDebugStream m(true,false,true);
        m << "sct_is_server named [" << server_name 
          << "] registering with SubscriptionManager.";
        m.flush();
        // As an sct_is_server (rather than an is_server) it is
        // now our duty to look for the SubscriptionManager ... 
        SubscriptionManagerModule::SubscriptionManager_var sm 
          = p.lookup<SubscriptionManagerModule::SubscriptionManager>
          (SubscriptionManagerModule::SubscriptionManager::instanceName);
        if (CORBA::is_nil(sm)) {
          // oh dear, it wasn't there ... what should be do?
          if (contIfSMNotFound) {
            // just carry on
          } else {
            Sct::MultiMessageDebugStream m(true,true,true);
            m << "Error: sct_is_server named [" << server_name << "] was forced to quit as there is no SubscriptionManager running in [" << partition_name <<"].  " << helpMessage;
            m.flush();
            return 1;
          };
        } else {
          /// NOTE THAT CALLING THE LINE BELOW WILL might cause the SubscriptionManager to call us back and make some immediate subscriptions.  Because (if this happens) we want it to succeed, it is important that BY THIS STAGE our Repository be fully functional.  I think that the Rep is indeed up at working at this point in the code as the only thing which follows this "if (lookForSM) block" is the png_synch and the Reposirtory->run() commands which are relatively inconsequenctial.
          sm->registerNewIsServer(server_name);
        };       
      } catch (...) {
        // oh dear, some kind of error registering with the SM
        if (contIfSMNotFound) {
          // ignore the error, whatever it was
        } else {
          std::string reason="Unknown";
          try {
            throw;
          } catch (const std::exception & e) {
            reason="std::exception: ";
            reason+=e.what();
          } catch (const CORBA::Exception & e) {
            reason="CORBA::Exception: ";
            reason+=e._name();
          } catch (...) {
            // Could not find out what the reason was.
          };
          Sct::MultiMessageDebugStream m(true,true,true);
          m << "Error: sct_is_server named [" << server_name << "] was forced to quit.  Reason for failure was [" << reason << "].  " << helpMessage;
          m.flush();
          return 1;
        };
      };
    };
    
    std::cout << OWLTime() << " :: the \"" << Repository -> name() 
			 << "\" IS server started in the partition \"" 
			 << Repository -> partition().name() << "\"" << std::endl;
    
    if(synchRequested) {
        pmg_initSync();
    };
    Repository -> run();
    
    if ( signalReceived != 0 )
    {
	std::cout << OWLTime() << "    :: [::main] the signal " 
		  << signalReceived << " received - stopping the \""
		  << Repository->name() << "\" IS server ... " ;
    }
    
    if ( alarm )
    {
	delete alarm;
    }
    
    backup( Repository );
    
    Repository->_destroy( true );
    
    std::cout << "done." << std::endl;
    
    return 0; 
}
